from fastchat.conversation import conv_templates, SeparatorStyle
from auto_vicuna.model import generate_stream


def main_loop(model, tokenizer, conv_template, temperature,
              max_new_tokens, plugins, debug, model_path):
    # Chat
    print(model)
    conv = conv_templates[conv_template].copy()
    while True:
        try:
            inp = input("User:")
        except EOFError:
            inp = ""
        if not inp:
            print("exit...")
            break

        conv.append_message(conv.roles[0], inp)
        conv.append_message(conv.roles[1], None)

        prompt = conv.get_prompt()
        skip_echo_len = len(prompt) + 1
        params = {
            "model": model_path,
            "prompt": prompt,
            "temperature": temperature,
            "max_new_tokens": max_new_tokens,
            "stop": conv.sep if conv.sep_style == SeparatorStyle.SINGLE else conv.sep2,
        }
        output_stream = generate_stream(
            model, tokenizer, params, model.device
        )

        pre = 0
        for outputs in output_stream:
            outputs = outputs[skip_echo_len:].strip()
            outputs = outputs.split(" ")
            now = len(outputs)
            if now - 1 > pre:
                print(" ".join(outputs[pre:now-1]), end=" ", flush=True)
                pre = now - 1

        print(" ".join(outputs[pre:now]), flush=True)
        output = " ".join(outputs)
        for plugin in plugins:
            output = plugin.on_response(output)
        conv.append_message(conv.roles[1], output)
        conv.offset += 2

    return conv



"""Handles loading of plugins."""

from ast import Module
import click
import zipfile
from pathlib import Path
from zipimport import zipimporter
from typing import List, Optional, Tuple


def inspect_zip_for_module(zip_path: str, debug: bool = False) -> Optional[str]:
    """
    Inspect a zipfile for a module.

    Args:
        zip_path (str): Path to the zipfile.
        debug (bool, optional): Enable debug logging. Defaults to False.

    Returns:
        Optional[str]: The name of the module if found, else None.
    """
    with zipfile.ZipFile(zip_path, 'r') as zfile:
        for name in zfile.namelist():
            if name.endswith("__init__.py"):
                if debug:
                    print(f"Found module '{name}' in the zipfile at: {name}")
                return name
    if debug:
        print(f"Module '__init__.py' not found in the zipfile @ {zip_path}.")
    return None


def scan_plugins(plugins_path: Path, debug: bool = False) -> List[Tuple[str, Path]]:
    """Scan the plugins directory for plugins.

    Args:
        plugins_path (Path): Path to the plugins directory.

    Returns:
        List[Path]: List of plugins.
    """
    plugins = []
    for plugin in plugins_path.glob("*.zip"):
        if module := inspect_zip_for_module(str(plugin), debug):
            plugins.append((module, plugin))
    return plugins


def load_plugins(plugins_path: Path, debug: bool = False) -> List[Module]:
    """Load plugins from the plugins directory.

    Args:
        plugins_path (Path): Path to the plugins directory.

    Returns:
        List[Path]: List of plugins.
    """
    plugins = scan_plugins(plugins_path)
    plugin_modules = []
    for module, plugin in plugins:
        plugin = Path(plugin)
        module = Path(module)
        if debug:
            click.echo(f"Plugin: {plugin} Module: {module}")
        zipped_package = zipimporter(plugin)
        zipped_module = zipped_package.load_module(str(module.parent))
        for key in dir(zipped_module):
            if key.startswith("__"):
                continue
            a_module = getattr(zipped_module, key)
            a_keys = dir(a_module)
            if '_abc_impl' in a_keys and \
                    a_module.__name__ != 'AutoVicunaPluginTemplate':
                plugin_modules.append(a_module)
    return plugin_modules



"""
    https://stackoverflow.com/questions/49100806/
    pylint-and-subprocess-run-returning-exit-status-28
"""
import subprocess

cmd = " pylint src\\**\\*"
try:
    subprocComplete = subprocess.run(
        cmd, shell=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    print(subprocComplete.stdout.decode("utf-8"))
except subprocess.CalledProcessError as err:
    print(err.output.decode("utf-8"))



"""
    Main entry point for the auto_vicuna package.
"""
import os
import sys
from pathlib import Path

import click
import torch
from dotenv import load_dotenv

from auto_vicuna.model import load_model
from auto_vicuna.loop import main_loop
from auto_vicuna.plugins import load_plugins

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
load_dotenv()


@click.command()
@click.option(
    "--vicuna_weights",
    type=click.Path(exists=True),
    default=lambda: os.environ.get("VICUNA_WEIGHTS", ""),
)
@click.option("--num_gpus", type=int, default=1)
@click.option("--device", type=str, default=DEVICE)
@click.option("--debug", is_flag=True)
@click.option("--load_8bit", is_flag=True)
@click.option(
    "--plugins_path", type=click.Path(exists=True),
    default=lambda: os.environ.get("PLUGINS_PATH", Path(os.getcwd()) / "plugins")
)
def main(
    vicuna_weights: Path,
    num_gpus: int = 1,
    device: str = DEVICE,
    debug: bool = False,
    load_8bit: bool = False,
    plugins_path: Path = Path(os.getcwd()) / "plugins",
) -> None:
    """Auto-Vicuna: A Python package for automatically generating Vicuna randomness."""
    click.echo(f"Auto-Vicuna\n===========\nVersion: 0.1.0\nWeights: {vicuna_weights}")
    click.echo(f"Device: {device}\nTorch version: {torch.__version__}")
    if "cpu" in torch.__version__:
        click.echo("\nError: CPU not supported. Install a GPU version of PyTorch.")
        click.echo("See https://pytorch.org/get-started/locally/ for more info.\n")
        sys.exit(1)
    try:
        model, tokenizer = load_model(
            vicuna_weights,
            device=device,
            num_gpus=num_gpus,
            debug=debug,
            load_8bit=load_8bit,
        )
    except ValueError as e:  # noqa: F841
        if "Tokenizer" in str(e):
            click.echo(
                "Error: Tokenizer not found. Likely due to transformers "
                "and your weights generated by different versions..."
            )
            click.echo(
                "Replace LLaMATokenizer with LlamaTokenizer in your configs"
                " or use the same version of transformers as your weights."
            )
            sys.exit(1)
        click.echo(
            "Error: CPU not supported. Select a GPU device instead, e.g."
            " --device=cuda"
        )
        sys.exit(1)
    plugins_found = load_plugins(plugins_path)
    loaded_plugins = [plugin() for plugin in plugins_found]
    if loaded_plugins:
        click.echo(f"\nPlugins found: {len(loaded_plugins)}\n"
                   "--------------------")
    for plugin in loaded_plugins:
        click.echo(f"{plugin._name}: {plugin._version} - {plugin._description}")

    if debug:
        click.echo(f"Model: {model}")

    main_loop(model, tokenizer, "bair_v1", temperature=0.7,
              max_new_tokens=512, plugins=loaded_plugins, debug=debug,
              model_path=vicuna_weights)


if __name__ == "__main__":
    main(sys.argv[1:])
